from contextlib import contextmanager

from stevedore import ExtensionManager

from medikit import steps
from medikit.pipeline import Pipeline
from medikit.settings import DEFAULT_FEATURES
from medikit.utils import format_file_content


class ConfigurationRegistry():
    def __init__(self):
        self._configs = {}
        self._features = {}
        self._pipelines = {}

        def register_feature(ext):
            self._features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    def __getitem__(self, item):
        return self._configs[item]

    def keys(self):
        return self._configs.keys()

    def require(self, *args):
        if len(args) == 1:
            return self._require(args[0])
        elif len(args) > 1:
            return tuple(map(self._require, args))
        raise ValueError('Empty.')

    @contextmanager
    def pipeline(self, name):
        if not name in self._pipelines:
            self._pipelines[name] = Pipeline()
        yield self._pipelines[name]

    @property
    def pipelines(self):
        return self._pipelines

    def _require(self, name):
        if not name in self._features:
            raise ValueError('Unknown feature {!r}.'.format(name))

        if name not in self._configs:
            self._configs[name] = self._features[name].Config()

        return self._configs[name]


def read_configuration(dispatcher, filename, variables, features, files):
    config = ConfigurationRegistry()

    # monkey patch placeholders
    import medikit
    _listen, _pipeline, _require = medikit.listen, medikit.pipeline, medikit.require
    medikit.listen, medikit.pipeline, medikit.require = dispatcher.listen, config.pipeline, config.require

    # todo, move this in a configurable place
    with medikit.pipeline('release') as release:
        release.add(steps.Install())
        release.add(steps.BumpVersion())
        release.add(steps.Make('update-requirements'))
        release.add(steps.Make('clean install'))  # test docs
        release.add(steps.System('git add -p .'))
        release.add(steps.Commit('Release: {version}', tag=True))

    # read configuration
    with open(filename) as f:
        # TODO use runpy?
        code = compile(f.read(), filename, 'exec')
    ctx = {'listen': dispatcher.listen}
    exec(code, ctx)

    # restore old values
    medikit.listen, medikit.pipeline, medikit.require = _listen, _pipeline, _require

    for k in variables.keys():
        if k in ctx:
            variables[k] = ctx[k]

    for feature in DEFAULT_FEATURES:
        config.require(feature)

    # old, deprecated way ...
    for feature in set(ctx.pop('enable_features', ())) - set(ctx.pop('disable_features', ())):
        config.require(feature)

    # current way ...
    features = features | set(config.keys())

    for k in files.keys():
        if k in ctx:
            files[k] = format_file_content(ctx[k])

    return variables, features, files, config


_all_features = {}


def load_features():
    if not _all_features:
        def register_feature(ext, all_features=_all_features):
            all_features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    return _all_features



"""
NodeJS / Yarn support.

"""

import json
import os

from medikit.events import subscribe
from medikit.feature import Feature, LAST_PRIORITY


class NodeJSFeature(Feature):
    requires = {'python', 'make'}

    @subscribe('medikit.feature.make.on_generate')
    def on_make_generate(self, event):
        event.makefile['YARN'] = '$(shell which yarn)'
        event.makefile['NODE'] = '$(shell which node)'

        event.makefile.get_target('install').install += [
            '$(YARN) install --production',
        ]

        event.makefile.get_target('install-dev').install += [
            '$(YARN) install',
        ]

    @subscribe('medikit.on_start')
    def on_start(self, event):
        package = {
            'name': event.config['python'].get('name'),
            'version': '0.0.0',
            'description': event.config['python'].get('description'),
            'author': event.config['python'].get('author'),
            'license': event.config['python'].get('license'),
        }

        self.render_file_inline(
            'package.json',
            json.dumps(package),
        )

    @subscribe('medikit.on_end', priority=LAST_PRIORITY)
    def on_end(self, event):
        os.system('yarn install')
        os.system('git add yarn.lock')


__feature__ = NodeJSFeature



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

from pip._vendor.distlib.util import parse_requirement

from medikit.events import subscribe
from . import Feature, SUPPORT_PRIORITY


class SphinxFeature(Feature):
    class Config(Feature.Config):
        def __init__(self):
            self._theme = None

        def set_theme(self, theme):
            self._theme = parse_requirement(theme)

        def get_theme(self):
            return self._theme

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(dev=[
            'sphinx >=1.6,<2.0',
        ])

        theme = event.config['sphinx'].get_theme()
        if theme:
            event.config['python'].add_requirements(dev=[theme.requirement])

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile

        makefile['SPHINX_BUILD'] = '$(PYTHON_DIRNAME)/sphinx-build'
        makefile['SPHINX_OPTIONS'] = ''
        makefile['SPHINX_SOURCEDIR'] = 'docs'
        makefile['SPHINX_BUILDDIR'] = '$(SPHINX_SOURCEDIR)/_build'

        makefile.add_target(
            '$(SPHINX_SOURCEDIR)',
            '''
            $(SPHINX_BUILD) -b html -D latex_paper_size=a4 $(SPHINX_OPTIONS) $(SPHINX_SOURCEDIR) $(SPHINX_BUILDDIR)/html
        ''',
            deps=('install-dev', ),
            phony=True
        )


__feature__ = SphinxFeature



from medikit.events import LoggingDispatcher
from medikit.feature import Feature
from medikit.file import NullFile
from unittest import TestCase


class FeatureTestCase(TestCase):
    feature_type = Feature

    def create_dispatcher(self):
        return LoggingDispatcher()

    def create_feature(self, feature_type=None, dispatcher=None):
        dispatcher = dispatcher or self.create_dispatcher()
        feature = (feature_type or self.feature_type)(dispatcher)
        feature.file_type = NullFile
        return feature, dispatcher
